#include <iostream>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

using namespace std;

bool isPrime(unsigned long number);
// Precondition:  computePrimes must have been invoked
//                number is a number to be tested for 
//                primality
//
// Postcondition: true - if number is prime
//                false - otherwise

void computePrime();
// Precondition:  none
//
// Postcondition: computes all primes between 0 and 1 billion
//                stored in a bit array "feld".  This function
//                takes a LONG time to execute, and should only
//                be called once at initialization of the process.

void *workerThread(void *clientSocket);
// Precondition:  clientSocket is a pointer to a live socket 
//                connection between a client and server
//
// Postcondition: handles communication/response with client

int main(void)
{
  // Compute the prime numbers
  computePrime();
  
  //////////////////////////////////////////////////////////////
  // Create a socket. 
  //////////////////////////////////////////////////////////////
  int s = socket(PF_INET,SOCK_STREAM,0);

  if (s <= 0) {
    cerr << "server: Socket creation failed." << endl;
    exit(-1);
  }

  /////////////////////////////////////////////////////////////
  // Create a port to listen to for connections
  /////////////////////////////////////////////////////////////
  sockaddr_in address;
  int addressSize = sizeof(sockaddr);
  address.sin_family=AF_INET;
  address.sin_addr.s_addr = INADDR_ANY;
  address.sin_port  = htons(15000);

  int status=0;
  status = bind(s,(struct sockaddr*) &address, addressSize);

  if (status != 0) {
    cerr << "server: Bind operation failed, unable to create listening port." << endl;
    exit(-1);
  }

  ////////////////////////////////////////////////////////////
  // DISPATCHER THREAD LOOP
  ////////////////////////////////////////////////////////////
  while (true) {

    // Wait for client connection
    cout << "server: Waiting for client to connect" << endl;
    listen(s,3);
    int cs = accept(s,(struct sockaddr*) &address,(socklen_t *) &addressSize);

    // Create a worker thread to handle prime number request
    pthread_t *tid = new pthread_t();
    pthread_create(tid,NULL,workerThread,&cs);

  }

  //
  // We should never get here... but if we did we need to close
  // the dispatcher's listener socket.
  //
  close(s);

  return 1;
}

void *workerThread(void *clientSocket)
{

  ////////////////////////////////////////////////////////////
  // Typecast and validate the socket.
  // 
  // Typecast the void pointer to an integer pointer and 
  // dereference the pointer to get at the integer that 
  // represents the client socket.
  //
  ////////////////////////////////////////////////////////////
  int cs = *((int *) clientSocket);

  if (cs <= 0) {
    cerr << "worker thread: accept() failed, client can't connect to port." << endl;
    cout << strerror(cs) << endl;
    exit(-1);
  }

  cout << "worker thread: Established communication with client." << endl;
  
  ////////////////////////////////////////////////////////////
  // Receive message from client, it is a BIG INTEGER
  ////////////////////////////////////////////////////////////
  unsigned long number = 0;
  recv(cs,&number,sizeof(long),0);

  ////////////////////////////////////////////////////////////
  // Determine if prime
  ////////////////////////////////////////////////////////////
  cout << "worker thread: Processing client request..." << endl;
  bool result = false;
  result = isPrime(number);

  ////////////////////////////////////////////////////////////
  // Send message to client, it is a small boolean
  ////////////////////////////////////////////////////////////
  send(cs,&result,sizeof(bool),0);
  cout << "worker thread: request complete" << endl;

  ////////////////////////////////////////////////////////////
  // Done!  Cleanup.
  ////////////////////////////////////////////////////////////
  cout << "worker thread: Closing connection with client." << endl;
  close(cs);
  pthread_exit(0);
}


//
// Stores the prime numbers in a lookup table
//
unsigned char *feld=NULL, *zzz;

// Don't worry about this syntax.  You aren't responsible
// for it.  If you're curious though... what you are looking
// at here are two MACRO definitions.  MACROs are a lot like
// functions in C++.
#define TEST(f,x)       (*(f+(x)/16)&(1<<(((x)%16L)/2)))
#define SET(f,x)        *(f+(x)/16)|=1<<(((x)%16L)/2)


bool isPrime(unsigned long number) {

  // Simulate some non-CPU blocking activity like I/O
  sleep(5);
  
  if (number==2) return true;

  if (number%2==0) return false;

  if (!TEST(feld,number)) return true;
}

void computePrime()
{
  unsigned long teste=1, max, mom, hits=1, count, alloc, s=0, e=1;
  time_t begin;

  //max = 4294967295L;
  max   = 100000L;
  
  while (feld==NULL)
  zzz = feld = (unsigned char *) malloc (alloc=(((max-=10000L)>>4)+1L));

  for (count=0; count<alloc; count++) *zzz++ = 0x00;

  cout << "server: Computing prime numbers to : " << max << endl;

  begin = time (NULL);
  while ((teste+=2) < max) {

    if (!TEST(feld, teste)) {
      if  (++hits%2000L==0) {
	cout << hits << "  prime number\x0d"; 
	cout.flush();
      }
    }

    for (mom=3L*teste; mom<max; mom+=teste<<1) SET (feld, mom);
  }

  cout << "server: " << hits << " prime numbers found " << time(NULL)-begin << " seconds." << endl;
}

